use std::ops::{Add, AddAssign, Sub, SubAssign};

use num::*;

pub mod strings;
pub use advent_of_code_macros::*;

pub fn matrix(zero: bool, diag: bool, non_diag: bool) -> Vec<(i32, i32)> {
    const MIN: i32 = -1;
    const MAX: i32 = 1;
    let mut d = Vec::new();
    for i in MIN..=MAX {
        for j in MIN..=MAX {
            let ia = i.abs();
            let ja = j.abs();
            if zero && i == 0 && j == 0 {
                d.push((i, j));
            }
            if diag && ia == 1 && ja == 1 {
                d.push((i, j));
            }
            if non_diag && ia != ja && ia + ja == 1 {
                d.push((i, j));
            }
        }
    }
    d
}

#[derive(Debug, Clone, Copy, Default, PartialEq, Eq, PartialOrd, Ord)]
pub struct Kartesian<T>
where
    T: Integer,
{
    pub x: T,
    pub y: T,
}

impl<T: Integer> Add for Kartesian<T> {
    fn add(self, rhs: Self) -> Self::Output {
        Kartesian {
            x: self.x + rhs.x,
            y: self.y + rhs.y,
        }
    }

    type Output = Kartesian<T>;
}

impl<T: Integer + Euclidian> AddAssign for Kartesian<T> {
    fn add_assign(&mut self, rhs: Self) {
        self.x += rhs.x;
        self.y += rhs.y;
    }
}

impl<T: Integer> Sub for Kartesian<T> {
    fn sub(self, rhs: Self) -> Self::Output {
        Kartesian {
            x: self.x - rhs.x,
            y: self.y - rhs.y,
        }
    }

    type Output = Kartesian<T>;
}

impl<T: Integer + Euclidian> SubAssign for Kartesian<T> {
    fn sub_assign(&mut self, rhs: Self) {
        self.x -= rhs.x;
        self.y -= rhs.y;
    }
}

impl<T: Integer> Kartesian<T> {
    pub const fn new(x: T, y: T) -> Self {
        Self { x: x, y: y }
    }
}

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub struct Euclidian<T>
where
    T: Integer,
{
    pub x: T,
    pub y: T,
    pub z: T,
}

impl<T: Integer> Add for Euclidian<T> {
    fn add(self, rhs: Self) -> Self::Output {
        Euclidian {
            x: self.x + rhs.x,
            y: self.y + rhs.y,
            z: self.z + rhs.z,
        }
    }

    type Output = Euclidian<T>;
}

impl<T: Integer + Euclidian> AddAssign for Euclidian<T> {
    fn add_assign(&mut self, rhs: Self) {
        self.x += rhs.x;
        self.y += rhs.y;
        self.z += rhs.z;
    }
}

impl<T: Integer> Sub for Euclidian<T> {
    fn sub(self, rhs: Self) -> Self::Output {
        Euclidian {
            x: self.x - rhs.x,
            y: self.y - rhs.y,
            z: self.z - rhs.z,
        }
    }

    type Output = Euclidian<T>;
}

impl<T: Integer + Euclidian> SubAssign for Euclidian<T> {
    fn sub_assign(&mut self, rhs: Self) {
        self.x -= rhs.x;
        self.y -= rhs.y;
        self.z -= rhs.z;
    }
}

impl<T: Integer> Euclidian<T> {
    pub const fn new(x: T, y: T) -> Self {
        Self { x: x, y: y, z: z }
    }
}

pub type Table = Vec<Vec<char>>;

#[derive(Debug, Clone, Copy, PartialEq, Eq, PartialOrd, Ord)]
pub enum KartesianDirection {
    TopLeft,
    Top,
    TopRight,
    Left,
    None,
    Right,
    BottomLeft,
    Bottom,
    BottomRight,
}

impl KartesianDirection {
    pub fn vector<T: Signed + Integer + Zero + From<i8>>(self) -> Kartesian<T> {
        const BELOW: i8 = -1;
        const ON: i8 = 0;
        const UPPER: i8 = 1;
        Kartesian::<T> {
            x: match self {
                Self::TopLeft | Self::Top | Self::TopRight => BELOW.into(),
                Self::Left | Self::None | Self::Right => ON.into(),
                Self::BottomLeft | Self::Bottom | Self::BottomRight => UPPER.into(),
            },
            y: match self {
                Self::TopLeft | Self::Left | Self::BottomLeft => BELOW.into(),
                Self::Top | Self::None | Self::Bottom => ON.into(),
                Self::TopRight | Self::Right | Self::BottomRight => UPPER.into(),
            },
        }
    }
}
